local function show_popup_panel(title, popup_text)
  local vb = renoise.ViewBuilder()
  renoise.app():show_custom_dialog(
  title, 
  vb:column {
    margin = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN,
    spacing = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING,
    uniform = true,
    vb:text {
      text = popup_text  
    } 
  }
  )
end

--------------------------------------------------------------------------------
-- GUI
--------------------------------------------------------------------------------

GUI = {}

GUI.dialog = nil
GUI.vb = {}
GUI.current_text = ""
GUI.lines_list = ""


GUI.values = nil
GUI.tool_name  = ""
GUI.tool_id   = ""
GUI.tool_version   = ""

GUI.probability_value_text = ""
GUI.start_line_text = "0"
GUI.func_text = ""
GUI.note_off_text = ""

GUI.spacing = 4 * renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
GUI.margin  = 1.5 * renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN

GUI.help_text = [[ 
The goal of this tool is to remove some percentage of notes from a note column.

There are two factors: a given probability, and a list of line numbers.

You need to enter the probablity as a number between 0 and 100.   
While there is little point in using 0%, it is allowed.

If you want to remove all notes that match a list of line numbers just set the probabilty to 100.

You can enter line numbers by hand, but you can also use a list generator function.

This generator understands a very simple and limited syntax:

`+ i j k` means line numbers increment by i, then j, then k, then i, then j, etc.


For example: 

`+ 3 5` would give you `0 3 8 11 16 19 24 ...` up to the number to lines in the current pattern.

`+ 3 2 5` would give you `0 3 5 10 13 15 20 23 25  ...` up to the number to lines in the current pattern.


`/ i j k` gives you all the lines evenly divisible by any of i, j, k, etc.

For example: 

`/ 3 5` would give you `0 3 5 6 9 10 12 15 18 20 21 ...` up to the number to lines in the current pattern.

`/ 6` would give you `0 6 12 18 24 30 ...` up to the number to lines in the current pattern.

Please note that there must be spaces between the characters in the function.

You can edit the resulting list.  Do not use commas in either the function or the line-number list.

Each time you run a line-generation function the generated numbers are *added* to the current list. 

This allows you to construct a more complex sequence by combining different line-generation functions.

There is a `clear` button to clear the current list; you can also hand-edit that list to fine-tune the values.

You must click the `Go` button in order to apply the results.

The last-applied values are saved and reloaded the next time the tool is used.

Note: Clearing the entire line number list will be treated as wanting to act on all lines.

A bit unintuitive, but having no lines is pointless, and it saves you the trouble of having to fill the list.


]]


-- ======================================================================
function line_nums_to_table(lines_str)
  print("In line_nums_to_table")
  local nums_table = string.int_list_to_numeric(lines_str)
  local paired_table = {}

  for k,v in pairs(nums_table ) do
    paired_table[v] = v
  end

  return paired_table 
end

-- ======================================================================
function load_values()
  print("Load last-used values ...")

  GUI.values = renoise.Document.create("ErodeLastValues") {
    gen_function        = GUI.func_text,
    lines_list          = GUI.lines_list,
    lines_negation      = GUI.lines_negation or false,
    note_off            = GUI.note_off or false,
    probability_value   = GUI.probability_value_text,
    start_line_text     = GUI.start_line_text

  }

  local ok, err = GUI.values:load_from("erode.xml")
  print("Loaded previous values. err = " , err)

  GUI.func_text               = GUI.values.gen_function.value 
  GUI.lines_list              = GUI.values.lines_list.value  
  GUI.lines_negation          = GUI.values.lines_negation.value
  GUI.note_off                = GUI.values.note_off.value
  GUI.probability_value_text  = GUI.values.probability_value.value
  GUI.start_line_text         = GUI.values.start_line_text.value

  print("Loaded values, GUI.func_text = " .. "'" .. GUI.func_text .. "'")

end

-- ======================================================================
function save_values()
  print("Save current values ...")

  if GUI.values ~= nil then
    GUI.values.gen_function.value       = GUI.func_text
    GUI.values.probability_value.value  = GUI.probability_value_text
    GUI.values.start_line_text.value    = GUI.start_line_text
    GUI.values.lines_list.value         = GUI.vb.views.lines_list_field.text
    GUI.values.lines_negation.value     = GUI.vb.views.lines_negation.value
    GUI.values.note_off.value           = GUI.vb.views.note_off.value

    GUI.values:save_as("erode.xml")
  end
end


-- ======================================================================
GUI.show_dialog = function()
  local LABEL_WIDTH = 110

  if GUI.dialog and GUI.dialog.visible then
    GUI.dialog:show()
    return
  else
    GUI.dialog = {}
  end

  GUI.vb = renoise.ViewBuilder()
  GUI.current_text = ""  

  load_values()

  local button_help = GUI.vb:button {
    text = "Help",
    width = 40,
    released = function()
      show_popup_panel("Help", GUI.help_text )
    end
  }


  local title = GUI.tool_name .. ". Version " .. GUI.tool_version

  local fx_prompt_row1 = GUI.vb:row {
    GUI.vb:text {
      text = "This is the Erode tool.\n" .. 
      "It clears notes in a given set of track-pattern lines based on a given probability\n\n"  .. 
      "You must have a note column selected for this to work.",
    },
  }

  local fx_prompt_row2 = GUI.vb:row {
    GUI.vb:text {
      text = "By default it simply clears the note, but there is an option to insert 'OFF' instead.\n" .. 
      "Line negation tells the tool to erode only lines that are not in the line list.\n\n" .. 
      "Click the Help button for more details.",
    },
  }

  local line_negation_row = GUI.vb:row {

    GUI.vb:checkbox {
      value = GUI.lines_negation or false,
      id = "lines_negation",
      notifier = function()
        print("GUI.vb.views.lines_negation clicked")
        GUI.lines_negation = GUI.vb.views.lines_negation.value
        print("Updated GUI.lines_negation: " , GUI.lines_negation) 
      end
    },

    GUI.vb:text {
      text = "Use negation of line numbers",
      tooltip = "Skip instead of apply-to",
    },

  }

  local note_off_row = GUI.vb:row {
    GUI.vb:checkbox {
      value = GUI.note_off or false,
      id = "note_off",
      notifier = function()
        print("GUI.vb.views.note_off clicked")
        GUI.note_off = GUI.vb.views.note_off.value
        print("Updated GUI.note_off: " , GUI.note_off) 
      end
    },

    GUI.vb:text {
      text = "Use note off instead of clearing the current value",
      tooltip = "Insert note off",
    },

  }

  local function_row_1 = GUI.vb:row {
    GUI.vb:text {
      width = LABEL_WIDTH,
      text = "Line selection function",
      tooltip = "Function to generate line numbers",
    },

    GUI.vb:textfield {
      text = GUI.func_text,
      id = "func_text",
      notifier = function()
        GUI.func_text = string.upper(GUI.vb.views.func_text.text)
        print("Updated GUI.func_text: " .. GUI.func_text)
      end,
    },


  } 

  local function_row_2 = GUI.vb:row {
    GUI.vb:text {
      text = "Start with this line (clear value to ignore)",
      width = LABEL_WIDTH,
      tooltip = "First line of generated list; 0 if left empty",
    },

    GUI.vb:textfield {
      text = GUI.start_line_text,
      width = 30,
      id = "start_line_text",
      notifier = function()
        GUI.start_line_text = string.upper(GUI.vb.views.start_line_text.text)
        print("Updated GUI.start_line_text: " .. GUI.start_line_text)
      end,
    },

  } 


  local function_row_3 = GUI.vb:row {
    GUI.vb:button {
      text = "Generate",
      released = function()
        GUI.generate_values()
      end
    } 
  } 

  local probability_value_row = GUI.vb:row {
    GUI.vb:text {
      width = LABEL_WIDTH,
      text = "Probability (0 to 100)",
      tooltip = "Likelihood of removal",
    },

    GUI.vb:textfield {
      text = GUI.probability_value_text,
      id = "probability_value_text",
      notifier = function()
        GUI.probability_value_text = string.upper(GUI.vb.views.probability_value_text.text)
        print("Updated GUI.probability_value_text: " .. GUI.probability_value_text)
      end,
    },

  } 

  local button_clear = GUI.vb:button {
    text = "Clear",
    released = function()
      GUI.vb.views.lines_list_field.text = ""
    end
  } 

  local button_go = GUI.vb:button {
    text = "Go",
    width = 40,
    released = function()
      Core.erode(GUI)
      save_values()
      GUI.dialog:close()
    end
  }

  local prompts_row = GUI.vb:row {
    GUI.vb:column {
      spacing = GUI.spacing,
      margin = GUI.margin,
      fx_prompt_row1,
      fx_prompt_row2,
    },
  }

  local bottom_buttons_row = GUI.vb:row {
    GUI.vb:horizontal_aligner {
      width = "100%",
      mode = "justify",
      spacing = 320,
      button_go, 
      button_help,
    },

  } 


  local content = GUI.vb:column {
    prompts_row, 
    GUI.vb:column {
      spacing = GUI.spacing,
      margin = GUI.margin,
      probability_value_row ,
      function_row_1,
      function_row_2,
      function_row_3,
    },

    GUI.vb:column {
      spacing = GUI.spacing,
      margin = GUI.margin,

      GUI.vb:text {
        text = "Apply to lines:",
        tooltip = "Series of line numbers",
      },

      GUI.vb:multiline_textfield {
        id = "lines_list_field",
        width = 400,
        height = 50,
        text = GUI.lines_list,
        notifier = function()
          GUI.lines_list = GUI.vb.views.lines_list_field.text
        end
      },

      GUI.vb:horizontal_aligner {
        mode = "right",
        button_clear,
      },


      line_negation_row,
      note_off_row,
      bottom_buttons_row,
    },

  } 

  GUI.dialog = renoise.app():show_custom_dialog(title, content)  

end


-- ======================================================================
GUI.generate_values = function() 

  local current_set = line_nums_to_table(GUI.vb.views.lines_list_field.text)
  local function_str  = string.trim(GUI.func_text)
  local start_line    = string.trim(GUI.start_line_text)

  if U.is_empty_string(start_line) then
    start_line = nil
  else
    start_line = tonumber(start_line)
  end

  if U.is_empty_string(function_str) then
    print("Function string is empty, returning.")
    return 
  end

  local new_set = Core.new_set_from_funct_string(function_str, current_set, start_line)

  print("New, generated, set: ")
  rprint(new_set)

  GUI.vb.views.lines_list_field.text = table.concat(new_set, " ")
  GUI.lines_list  = GUI.vb.views.lines_list_field.text
end

return GUI

